# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: TimeFilterFastResponse.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1cTimeFilterFastResponse.proto\x12*com.igeolise.traveltime.rabbitmq.responses\"\xbf\x06\n\x16TimeFilterFastResponse\x12W\n\x05\x65rror\x18\x01 \x01(\x0b\x32H.com.igeolise.traveltime.rabbitmq.responses.TimeFilterFastResponse.Error\x12\x61\n\nproperties\x18\x02 \x01(\x0b\x32M.com.igeolise.traveltime.rabbitmq.responses.TimeFilterFastResponse.Properties\x1aJ\n\nProperties\x12\x13\n\x0btravelTimes\x18\x01 \x03(\x11\x12\x14\n\x0cmonthlyFares\x18\x02 \x03(\x05\x12\x11\n\tdistances\x18\x03 \x03(\x05\x1a\x63\n\x05\x45rror\x12Z\n\x04type\x18\x01 \x01(\x0e\x32L.com.igeolise.traveltime.rabbitmq.responses.TimeFilterFastResponse.ErrorType\"\xb7\x03\n\tErrorType\x12\x0b\n\x07UNKNOWN\x10\x00\x12 \n\x1cONE_TO_MANY_MUST_NOT_BE_NULL\x10\x01\x12\x1b\n\x17SOURCE_MUST_NOT_BE_NULL\x10\x02\x12#\n\x1fTRANSPORTATION_MUST_NOT_BE_NULL\x10\x03\x12\x1a\n\x16SOURCE_NOT_IN_GEOMETRY\x10\x04\x12$\n UNRECOGNIZED_TRANSPORTATION_MODE\x10\x05\x12\x1d\n\x19TRAVEL_TIME_LIMIT_TOO_LOW\x10\x06\x12\x1e\n\x1aTRAVEL_TIME_LIMIT_TOO_HIGH\x10\x07\x12\x19\n\x15\x41UTH_ERROR_NO_USER_ID\x10\x08\x12.\n*SERVICE_MISMATCH_WRONG_TRANSPORTATION_MODE\x10\t\x12\x17\n\x13SOURCE_OUT_OF_REACH\x10\n\x12/\n+INTERLEAVED_DELTAS_INVALID_COORDINATE_PAIRS\x10\x0b\x12#\n\x1f\x44ISTANCE_PROPERTY_NOT_SUPPORTED\x10\x0c\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'TimeFilterFastResponse_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _TIMEFILTERFASTRESPONSE._serialized_start=77
  _TIMEFILTERFASTRESPONSE._serialized_end=908
  _TIMEFILTERFASTRESPONSE_PROPERTIES._serialized_start=291
  _TIMEFILTERFASTRESPONSE_PROPERTIES._serialized_end=365
  _TIMEFILTERFASTRESPONSE_ERROR._serialized_start=367
  _TIMEFILTERFASTRESPONSE_ERROR._serialized_end=466
  _TIMEFILTERFASTRESPONSE_ERRORTYPE._serialized_start=469
  _TIMEFILTERFASTRESPONSE_ERRORTYPE._serialized_end=908
# @@protoc_insertion_point(module_scope)
